"""
Title: 
Author: Julio Rodriguez
Organization: SENER
"""

# BLK: Imports

# Standard Libraries
import cProfile
import pstats


# 3rd Party
# import pandas as pd
import numpy as np


# Local Libraries


# Global variables
class Cstruct:
    pass


# Code

# Clases

# Funciones
def describe(dataframe):
    """
    Soluciona el problema con PyCharm al utilizar la función describe() que pone % en los índices
    :param dataframe: Pandas DataFrame a aplicar la función describe
    :return: Función describe aplicada sin % en los índices
    """
    df_describe = dataframe.describe().rename(index={'25%': '25', '50%': '50', '75%': 75})
    return df_describe


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]


def rms_calculate(array):
    return np.sqrt(np.mean(array ** 2))



"""
Title: 
Author: Julio Rodriguez
Organization: SENER
"""

# BLK: Imports

# Standard Libraries
from datetime import datetime

# 3rd Party
import numpy as np
import pandas as pd
from tqdm import tqdm
import openpyxl as opx

# Local Libraries
import solar_general as _gn


# Global class
class Cstruct:
    pass


# Global variables


# Code


def inicializa_deltas(DLT, datos_tmpt, kks_loops, n_locs=4):
    # FuncBLK: inicializa_deltas
    DLT.datos_tmpt_ordenados = np.zeros([datos_tmpt.shape[1], kks_loops.size, n_locs], dtype=float)     # Temperaturas de cada LOC en cada muestra ordenadas por lazo y en orden de 1-4 LOCs
    DLT.deltas = np.zeros([datos_tmpt.shape[1], kks_loops.size, n_locs], dtype=float)   # Incremento de temperatura de un LOC a otro (primer delta es de Tin a LOC1)
    DLT.dtns = np.zeros([datos_tmpt.shape[1], kks_loops.size, n_locs], dtype=float)     # Incremento de temperatura estimado desde el principio del panel hasta su LOC (en el medio)
    DLT.paneles = np.zeros([datos_tmpt.shape[1], kks_loops.size, n_locs], dtype=float)  # Estimación de lo que calienta un panel utilizando dtns. Inestable cuando una PT100 no mide bien.
    DLT.delta_tot = np.empty([datos_tmpt.shape[1], kks_loops.size], dtype=float)        # Incremento de temperatura total estimado
    DLT.delta_prom = np.empty([datos_tmpt.shape[1], kks_loops.size], dtype=float)       # Incremento medio estimado de temperatura (dtns prom.)
    DLT.delta_desv = np.empty([datos_tmpt.shape[1], kks_loops.size], dtype=float)       # Desviación típica de el incremento de temperatura estimado(std dtns)
    DLT.t_out_est = np.empty([datos_tmpt.shape[1], kks_loops.size], dtype=float)        # Temperatura de salida estimada
    DLT.t_in = np.empty([datos_tmpt.shape[1], kks_loops.size], dtype=float)             # Temperatura de entrada (dato)
    DLT.t_out = np.empty([datos_tmpt.shape[1], kks_loops.size], dtype=float)            # Temperatura de salida (dato)
    return DLT


def genera_deltas(kks_loops, index_kks_loops, pos_loc, datos_tmpt, datos_general, med, DLT, sect_div, pt100_est, pt100_d, is_pt_sum, ct_pos, ht_pos, n_locs=4):   # Si noche es True no hay que pasar la pt100_noche
    sec = 0
    # Si la variable is_pt_sum es cierta, entonces se hace que pt100_d sea 0 para no sumar la diferencia estimada de arreglo de PT100
    if is_pt_sum:
        pt100_d = np.zeros(pt100_d.shape, dtype=float)
    for k in np.arange(kks_loops.size):  # Loop de cada loop(kksloop)
        if sect_div[sec + 1] < k:  # Comprobar en que sector estamos (plantilla ordenada por sectores)
            sec += 1

        # Ordenar por nº de LOC
        pos = np.argwhere(index_kks_loops == k).squeeze()  # Posiciones de lás máquinas del loop actual
        pre_pos_ordenado = np.argsort(pos_loc[pos] - 1)  # Sacar las posiciones del array ordenado de las posiciones de cada LOC (mejor mirarlo en debugging)
        pos_ordenado = pos[pre_pos_ordenado]  # Posiciones de los LOCs ordenados por numero de LOC
        DLT.datos_tmpt_ordenados[med, k, 0] = datos_tmpt[pos_ordenado[0], med] - pt100_d[k, 0]

        # Primer LOC
        DLT.deltas[med, k, 0] = DLT.datos_tmpt_ordenados[med, k, 0] - datos_general[ct_pos + sec, med]  # Diferencia (en temperatura) entre el primer LOC y la temperatura de entrada
        DLT.dtns[med, k, 0] = DLT.datos_tmpt_ordenados[med, k, 0] - datos_general[ct_pos + sec, med]
        DLT.paneles[med, k, 0] = 2 * DLT.dtns[med, k, 0]

        # Resto de LOCs
        for i in np.arange(n_locs - 1) + 1:
            DLT.datos_tmpt_ordenados[med, k, i] = datos_tmpt[pos_ordenado[i], med] - pt100_d[k, i]
            DLT.deltas[med, k, i] = DLT.datos_tmpt_ordenados[med, k, i] - DLT.datos_tmpt_ordenados[med, k, i - 1] # Diferencia del LOC i y el i - 1
            DLT.dtns[med, k, i] = DLT.deltas[med, k, i] - DLT.dtns[med, k, i - 1]
            DLT.paneles[med, k, i] = 2 * DLT.dtns[med, k, i]

        # Calculos
        DLT.delta_tot[med, k] = DLT.datos_tmpt_ordenados[med, k, -1] - datos_general[ct_pos + sec, med]  # Diferencia entre el último LOC y la temperatura de entrada
        DLT.delta_prom[med, k] = (DLT.delta_tot[med, k]) / 7  # Delta total entre 7 nos da la media de cuanto vale un dtn aprox.
        DLT.delta_desv[med, k] = np.std(DLT.dtns[med, k, :])  # Desviación típica de la delta estimada
        DLT.t_out_est[med, k] = DLT.datos_tmpt_ordenados[med, k, 3] + DLT.dtns[med, k, 3]   # Temperatura de salida estimada
        DLT.t_in[med, k] = datos_general[ct_pos + sec, med]   # T entrada sector
        DLT.t_out[med, k] = datos_general[ht_pos + sec, med]  # T salida sector

    return DLT


def datos_loop_a_loc(kks_loops_full, kks_loops, *args):
    out = np.empty(0)
    args = list(args)
    try:
        temp = args[0].shape[1]
    except IndexError:
        args[0] = args[0][:, np.newaxis]
    for data in args:
        try:
            temp = np.empty([kks_loops_full.size, data.shape[1]], dtype=data.dtype)
        except IndexError:
            data = data[:, np.newaxis]
            temp = np.empty([kks_loops_full.size, data.shape[1]], dtype=data.dtype)
        for k in np.arange(kks_loops.size):
            if data.shape[1] == 4:
                pos_locs = np.argwhere(kks_loops_full == kks_loops[k])
                for i in np.arange(pos_locs.size):
                    temp[pos_locs[i]] = data[k]
            else:
                pos_locs = np.argwhere(kks_loops_full == kks_loops[k])
                for i in np.arange(pos_locs.size):
                    temp[pos_locs[i, 0]] = data[k]
        if np.array_equal(data, args[0]):
            out = temp
        else:
            out = np.append(out, temp, axis=1)
    return out


def data_loc4_to_loc(data, index_kks_loops, pos_loc):
    data_out = np.empty(0)
    for k in np.arange(index_kks_loops.size):
        # Conseguir que se sincronice el 4loc a loc
        data_out = np.append(data_out, data[index_kks_loops[k], pos_loc[k] - 1])
    return data_out


def array_division(datos):
    comp_arr = np.zeros(datos.size, dtype=int)  # Crea un array de ceros del tipo integer del mismo tamaño que fechas
    k1 = 0
    div_arr = np.empty(0, dtype=int)  # Crea un array vacio del tipo integer sin nada dentro
    while comp_arr[-1] != (datos.size - 1):  # Con este loop se busca conseguir un array con las posiciones de la última alarma de cada día para poder dividir la lista de alarmas que está ordenada cronológicamente
        comp = datos[k1]  # Se le da el valor de la fecha a comparar en cada vuelta de loop
        comp_arr = np.argwhere(datos == comp)  # Esta función nos devuelve un array con los índices donde el array de fechas coincide con la fecha actual del loop
        k1 = comp_arr[-1] + 1  # Se actualiza k1 con el último índice del array de los índices
        div_arr = np.append(div_arr, comp_arr[-1], axis=0)  # Se guarda el último valor del índice para tener de referencia hasta donde dura cada día
    return div_arr


def corregir_texto_alarmas(origin):
    inst = np.array(origin, dtype='U140')
    inst = np.char.replace(inst, 'Loc ', '')
    inst = np.core.defchararray.rstrip(inst)  # Quita los espacios al final
    return inst


def corregir_texto_alarmas_V2(origin):
    inst = np.char.replace(origin, 'Loc ', '')
    inst = np.char.replace(inst, '/', '-')
    inst = np.char.replace(inst, 'HE', 'GX')
    inst = np.core.defchararray.rstrip(inst)
    return inst


def convertir_df_numerico_GENERAL(df, nombres_columnas_general, planta):    # FIXME: NO GLOBAL
    if planta == "Ilanga":
        df = _gn.convertir_df_numerico(df, [nombres_columnas_general[0], nombres_columnas_general[3], nombres_columnas_general[4], nombres_columnas_general[5]])
    elif planta == "Kathu":
        df = _gn.convertir_df_numerico(df, [nombres_columnas_general[0], nombres_columnas_general[4], nombres_columnas_general[5], nombres_columnas_general[9], nombres_columnas_general[12]])
    elif planta == "NoorII":
        df = _gn.convertir_df_numerico(df, ["ID", "Latitude (º)", "Longitude (º)", "RING / SUBFIELD", "LOC POSITION IN THE LOOP"])
    elif planta == "NoorIII":
        df = _gn.convertir_df_numerico(df, [nombres_columnas_general[0], nombres_columnas_general[3], nombres_columnas_general[4], nombres_columnas_general[12]])
    return df


def generar_excel_planta(Data, planta, datos, col_datos, cols_numeric=[], fich_nom=None, hora_select=None):
    # BLK: Generar fichero Excel

    if hora_select is None:
        hora_selected = Data.fecha_archivo
    else:
        hora_selected = hora_select.replace(":", "-")

    if fich_nom is None:
        fich_nom = Data.ruta_defecto + f"/{planta}" + "/Conclusiones " + Data.fecha_archivo + f"/Conclusiones_{hora_selected}.xlsx"

    print("Generando fichero Excel...")
    colgen_tam = Data.nombres_columna_general.size
    # sBLK: Creación del fichero
    writer = pd.ExcelWriter(fich_nom, engine='xlsxwriter')
    # Generar dataframes de pandas con los arrays de numpy para el volcado a excel
    df_temp = pd.DataFrame(datos, columns=col_datos)
    # Convertir celdas de excel al formato numérico (sin formato por defecto)
    df_temp = convertir_df_numerico_GENERAL(df_temp, Data.nombres_columna_general, planta)
    df_temp = _gn.convertir_df_numerico(df_temp, col_datos[[colgen_tam + elem for elem in cols_numeric]])  # Cal <-> DIF
    # Volcar dataframes a excel
    df_temp.to_excel(writer, index=False, header=True, sheet_name="Conclusiones")
    # Ajuste del ancho de las columnas
    _gn.ajustar_columnas_excel(writer, "Conclusiones", len(col_datos), df_temp.columns, normal=True)

    # Activar autofilter
    worksheet = writer.sheets["Conclusiones"]
    worksheet.autofilter(0, 0, Data.kks_loops_full.size, col_datos.size)

    # Guardar Excel
    writer.save()

    # sBLK: Modificar características del Excel
    # Inmovilizar primera fila
    excel_planta = opx.load_workbook(fich_nom)
    nom_hojas = excel_planta.sheetnames
    hoja_conclusiones = excel_planta["Conclusiones"]
    hoja_conclusiones.freeze_panes = "A2"
    excel_planta.save(fich_nom)


def ordenar_locs(datos_tmpt, kks_loops, index_kks_loops, pos_loc, n_locs=4):
    datos_tmpt_ordenados = np.zeros([datos_tmpt.shape[1], kks_loops.size, n_locs], dtype=float)  # Temperaturas de cada LOC en cada muestra ordenadas por lazo y en orden de 1-4 LOCs
    for med in np.arange(datos_tmpt.shape[1]):  # Loop de cada medición horaria
        # Ordenar por nº de LOC
        for k in np.arange(kks_loops.size):  # Loop de cada loop(kksloop)
            pos = np.argwhere(index_kks_loops == k).squeeze()  # Posiciones de lás máquinas del loop actual
            pre_pos_ordenado = np.argsort(pos_loc[pos] - 1)  # Sacar las posiciones del array ordenado de las posiciones de cada LOC (mejor mirarlo en debugging)
            pos_ordenado = pos[pre_pos_ordenado]  # Posiciones de los LOCs ordenados por numero de LOC
            datos_tmpt_ordenados[med, k, 0] = datos_tmpt[pos_ordenado[0], med]
            for i in np.arange(n_locs - 1) + 1:
                datos_tmpt_ordenados[med, k, i] = datos_tmpt[pos_ordenado[i], med]
    return datos_tmpt_ordenados


def crea_carpetas_conclusiones(Data, planta):
    _gn.crear_directorio_v2(Data.ruta_defecto, f"{planta}")
    return _gn.crear_directorio_v2(Data.ruta_defecto, f"{planta}/Conclusiones {Data.fecha_archivo}")


def eliminar_alarmas(alarmas, tipos_alarma, fechas_alarmas, alarmas_deseadas):
    index_particion = np.ones(alarmas.size, dtype=bool)     # Array de unos con el tamaño de las alarmas originales
    for i in tqdm(np.arange(alarmas_deseadas.size)):
        comp = np.argwhere(alarmas == alarmas_deseadas[i])  # Obtener indices de las posiciones de las alarmas que nos interesan
        if comp.any():
            np.put(index_particion, comp, False)    # Poner False en dichas posiciones de las alarmas originales
    index_particion = np.argwhere(index_particion)  # Obtener las posiciones donde todavía hay True en las alarmas originales, es decir, las que no nos interesan
    alarmas_new = np.delete(alarmas, index_particion, None)     # Eliminar dichas alarmas (inst)
    tipos_alarma_new = np.delete(tipos_alarma, index_particion, None)   # Eliminar los tipos de alarmas (desc)
    fechas_alarmas_new = np.delete(fechas_alarmas, index_particion, None)   # Eliminar las fechas (fechas)
    return alarmas_new, tipos_alarma_new, fechas_alarmas_new


def array_division_dias_csvalarm(fechas, planta):
    fecha = np.empty(fechas.size, datetime)                             # Crea un array vacío de tipo datetime del mismo tamaño que fechas
    for k in np.arange(fechas.size):                                    # Transformamos las fechas de tipo string a tipo datetime (para poder quitar la parte de la hora y poder así filtrar por día las alarmas)
        if planta != "NoorIII":
            fecha[k] = datetime.strptime(fechas[k], '%d/%m/%Y %H:%M:%S')    # Se le indica a la función que formato de fecha y hora tenemos en nuestro archivo
        else:
            fecha[k] = datetime.strptime(fechas[k], '%d/%m/%Y %H:%M')    # Es necesario hacer este if ya que el csv de NoorIII sale con la hora en diferente formato
        fecha[k] = np.datetime64(fecha[k], 'D')                         # Nos quedamos con la parte de fecha solo (AAAA-MM-DD)

    comp_arr = np.zeros(fechas.size, dtype=int)                         # Crea un array de ceros del tipo integer del mismo tamaño que fechas
    k1 = 0
    div_arr = np.empty(0, dtype=int)                                    # Crea un array vacio del tipo integer sin nada dentro
    while comp_arr[-1] != (fechas.size - 1):                            # Con este loop se busca conseguir un array con las posiciones de la última alarma de cada día para poder dividir la lista de alarmas que está ordenada cronológicamente
        comp = fecha[k1]                                                # Se le da el valor de la fecha a comparar en cada vuelta de loop
        comp_arr = np.argwhere(fecha == comp)                           # Esta función nos devuelve un array con los índices donde el array de fechas coincide con la fecha actual del loop
        k1 = comp_arr[-1] + 1                                           # Se actualiza k1 con el último índice del array de los índices
        div_arr = np.append(div_arr, comp_arr[-1], axis=0)              # Se guarda el último valor del índice para tener de referencia hasta donde dura cada día
    return div_arr, fecha                                                      # Este loop se repite hasta que en la última posición del array de los indices aparezca el último indice de la lista de alarmas


def init_dir(ruta_defecto, plant, fecha_archivo):
    # Directorios mínimos
    _gn.crear_directorio_v2(ruta_defecto, plant)
    _gn.crear_directorio_v2(ruta_defecto, plant + "/Conclusiones " + fecha_archivo)




"""
Title:
Author: Julio Rodriguez
Organization: SENER
"""

# BLK: Imports

# Standard Libraries
from datetime import datetime, timedelta
import sys
import time
import os
import io
from multiprocessing import Process, Pool, Semaphore, cpu_count
import subprocess
from pathlib import Path
from functools import wraps
import pickle
import cProfile
import pstats


# 3rd Party
# import pandas as pd
import numpy as np
import openpyxl.utils.exceptions
import pandas as pd
from tqdm import tqdm
from openpyxl import Workbook


# Local Libraries
import solar_common as _cm


# Global variables
class Cstruct:
    pass


# Code

# Clases

# Funciones


def crear_directorio_v2(ruta_defecto, nombre):
    """
    A diferencia de la versión 1, esta función no necesita que se le pase el nombre con la barra '/' delante, realiza un join internamente.

    :param ruta_defecto: string
    Ruta base donde se va a crear el directorio

    :param nombre: string
    Nombre del directorio nuevo

    :return: string
    Devuelve la ruta creada

    """
    ruta_final = os.path.join(ruta_defecto, nombre)
    if not os.path.exists(ruta_final):
        os.mkdir(ruta_final)
    return ruta_final


def porcentaje_x(array, tam, axis=0):
    array = np.true_divide(np.sum(array, axis=axis), tam) * 100
    return array


def consecutive(data, stepsize=1):
    return np.split(data, np.where(np.diff(data) != stepsize)[0] + 1)


def convert_to_float(frac_str):
    '''
    Convierte un string que contiene una fracción en numero decimal con el que operar en Python.
    :param frac_str: String que contiene la fracción.
    :return: Fracción convertida en numero de tipo float.
    '''
    try:
        return float(frac_str)
    except ValueError:
        num, denom = frac_str.split('/')
        try:
            num_temp = num
            leading, num = num_temp.split(' ')
            whole = float(leading)
        except ValueError:
            try:
                float(num)
                whole = 0
            except ValueError:
                if num_temp[0] == '-':
                    temp, leading, num = num_temp.split(' ')
                    whole = -float(leading)
                    num = '-' + num

        frac = float(num) / float(denom)
        return whole - frac if whole < 0 else whole + frac


def estimar_datos(datos, num_datos, tolerancia):
    suma = np.true_divide(np.sum(datos, axis=0), num_datos)
    estimacion = np.where(suma > tolerancia, 1, 0)
    return estimacion


def appendear_datos(first_data, axis, *args):
    try:
        first_data.shape[axis]
    except IndexError:
        first_data = first_data[:, np.newaxis]
    finally:
        out = first_data
    for data in args:
        try:
            data.shape[axis]
        except IndexError:
            try:
                data = data[:, np.newaxis]
            except IndexError:
                data = [data]
        finally:
            out = np.append(out, data, axis=axis)
    return out


def obtener_x_maxmin(data, n_elementos=None, orden='max'):
    """
    Obtiene los n_elementos máximos (max) o mínimos (min) de un ndarray.

    :param data: Datos a ser ordenados (ndarray)
    :param n_elementos: Nº máx. de elementos a devolver
    :param orden: De menor a mayor(min) o de mayor a menor(max)
    :return: Array data ordenado y array con los indices de ordenado
    """
    if n_elementos is None:
        n_elementos = data.size
    # Obtener los peores lazos (los que tengan mayor valor o menor dependiendo de abrir (negativo) o cerrar (positivo))
    # ind_tmp = np.argpartition(data, -n_elementos)[-n_elementos:]
    if orden == 'min':
        # ind_tmp_ord = ind_tmp[np.argsort(data[ind_tmp])]
        ind_tmp_ord = np.argsort(data)[:n_elementos]
    elif orden == 'max':
        # ind_tmp_ord = np.flip(ind_tmp[np.argsort(data[ind_tmp])])
        ind_tmp_ord = np.argsort(data)[-n_elementos:]
    else:
        raise ValueError("El valor introducido no es aceptado. Utiliza 'min' para obtener los mínimos y 'max' para obtener los máximos")

    return data[ind_tmp_ord], ind_tmp_ord


def multiprocesamiento(function, n_procesos, max_procesos=cpu_count(), *args):
    sema = Semaphore(max_procesos)
    all_processes = []
    for i in range(n_procesos):
        # once 20 processes are running, the following `acquire` call
        # will block the main process since `sema` has been reduced
        # to 0. This loop will continue only after one or more
        # previously created processes complete.
        sema.acquire()
        p = Process(target=function, args=(args, sema))
        all_processes.append(p)
        p.start()

    # inside main process, wait for all processes to finish
    for p in all_processes:
        p.join()


def generar_video_mp4(fig, fps, update, nom_video, n_frames, *args, **kwargs):
    # Inicialización de kwargs
    kwargs["returns"] = kwargs.get("returns", None)

    if type(fps) != str:
        fps = str(fps)

    canvas_width, canvas_height = fig.canvas.get_width_height()

    # Open an ffmpeg proce
    outf = nom_video
    cmdstring = ('ffmpeg',
                 '-y', '-r', fps,  # overwrite, x fps
                 '-s', '%dx%d' % (canvas_width, canvas_height),  # size of image string
                 '-pix_fmt', 'argb',  # format
                 '-f', 'rawvideo', '-i', '-',  # tell ffmpeg to expect raw video from the pipe
                 '-hide_banner', '-loglevel', 'error',   # Hides the description of the options selected
                 '-vcodec', 'mpeg4',
                 '-b:v', '5000k', outf)  # output encoding
    p = subprocess.Popen(cmdstring, stdin=subprocess.PIPE, shell=True)

    # Draw 1000 frames and write to the pipe
    for frame in tqdm(np.arange(n_frames)):

        # draw the frame, PONER AQUI LA FUNCIÓN QUE DIBUJA
        kwargs["returns"] = update(fig, frame, *args, **kwargs)
        fig.canvas.draw()

        # extract the image as an ARGB string
        string = fig.canvas.tostring_argb()

        # write to pipe
        p.stdin.write(string)

        # Clear the figure
        fig.clf()

    # Finish up
    p.communicate()
    return kwargs["returns"]


def convertir_df_numerico(df, columna):
    df[columna] = df[columna].apply(pd.to_numeric)
    return df


def ajustar_columnas_excel(writer, nombre_hoja, n_col, data, index=None, sectores_extendido=False, resumen=False, normal=False, fila=0, col_off=0):
    # Ajusta el ancho de las columnas de un archivo excel
    worksheet = writer.sheets[nombre_hoja]
    # width = np.max([np.max([len(index.values)]), len(data.names[0])]) + 2
    if normal:
        for i in np.arange(n_col):
            width = len(data.values[i]) + 2
            worksheet.set_column(fila, i + col_off, width)
    else:
        width = len(data.names[0]) + 2  # Primera columna
        worksheet.set_column(0, 0, width)
        if sectores_extendido:
            width = np.max([len(data.values[0][0]), len(data.values[0][1])]) + 2  # Máximo entre la fecha y la alarma
            worksheet.set_column(1, 1, width)
            width = np.max([len(data.values[0][0]), len(data.values[1][1])]) + 2  # Máximo entre la fecha y la máquina
            worksheet.set_column(2, 2, width)
            for i in np.arange(2, n_col, 2):
                width1 = len(data.values[i][1]) + 2
                width2 = len(data.values[i + 1][1]) + 2
                worksheet.set_column(i + 1, i + 1, width1)
                worksheet.set_column(i + 2, i + 2, width2)
        elif resumen:
            for i in np.arange(n_col - 1) + 1:  # El +1 -1 es porque el primer elemento lo meto arriba como names, pero podría pasarse de el y poner sin -1 +1.
                try:
                    if data.values[i][0] != data.values[i + 1][0] and data.values[i][0] != data.values[i - 1][0]:  # Este if mira a ver si seguimos en la misma multicolumna, y uso el try porque en la última va a dar un error de out of index
                        width = len(data.values[i][0]) + 2
                    else:
                        width = len(data.values[i][1]) + 2
                except:
                    if i == 0:  # Este error no debería saltar ya que el for empieza con el primer elemento = 1
                        if data.values[i][0] != data.values[i + 1][0]:
                            width = len(data.values[i][0]) + 2
                        else:
                            width = len(data.values[i][1]) + 2
                    elif i == n_col - 1:  # Este error salta cuando llega a la última posicion el for ya que no existe última + 1, por lo que la realiza dentro de esta excepción
                        if data.values[i][0] != data.values[i - 1][0]:
                            width = len(data.values[i][0]) + 2
                        else:
                            width = len(data.values[i][1]) + 2
                finally:
                    worksheet.set_column(i, i, width)

        else:  # Cuando no tiene ningun multiindex, hace un ajuste normal
            for i in np.arange(n_col) + 1:
                width = len(data.values[i - 1][0]) + 4
                worksheet.set_column(i, i, width)


def string_to_time(a_tiempo_str):
    t_alarma_tot = np.full(a_tiempo_str.size, timedelta(days=0, hours=0, minutes=0, seconds=0), dtype=timedelta)
    a_tiempo_str = np.char.replace(a_tiempo_str, "/", "-")
    try:
        datetime.strptime(a_tiempo_str[0], '%H:%M:%S')
    except ValueError:
        try:
            datetime.strptime(a_tiempo_str[0], '%Y-%m-%d %H:%M:%S')
        except:
            for k in np.arange(a_tiempo_str.size):
                tiempo_tmp = datetime.strptime(a_tiempo_str[k], '%d-%m-%Y %H:%M:%S')
                t_alarma_tot[k] = t_alarma_tot[k] + timedelta(hours=tiempo_tmp.hour, minutes=tiempo_tmp.minute, seconds=tiempo_tmp.second)
        else:
            for k in np.arange(a_tiempo_str.size):
                tiempo_tmp = datetime.strptime(a_tiempo_str[k], '%Y-%m-%d %H:%M:%S')
                t_alarma_tot[k] = t_alarma_tot[k] + timedelta(hours=tiempo_tmp.hour, minutes=tiempo_tmp.minute, seconds=tiempo_tmp.second)
    else:
        for k in np.arange(a_tiempo_str.size):
            tiempo_tmp = datetime.strptime(a_tiempo_str[k], '%H:%M:%S')
            t_alarma_tot[k] = t_alarma_tot[k] + timedelta(hours=tiempo_tmp.hour, minutes=tiempo_tmp.minute, seconds=tiempo_tmp.second)
    return t_alarma_tot


def find_data_file(filename):
    if getattr(sys, 'frozen', False):
        # The application is frozen
        datadir = os.path.dirname(sys.executable)
    else:
        # The application is not frozen
        # Change this bit to match where you store your data files:
        datadir = os.path.dirname(__file__)
    return os.path.join(datadir, filename)


def read_corrupt(filename, sheet_name='Hoja1', skiprows=0, df_return=False):

    # Nombre fichero de salida
    # Se realiza un split para añadir _FIX al final del nombre del archivo
    f_split = filename.split('.')
    # Crear subcarpeta para meter ahí los ficheros modificados
    crear_directorio_v2(str(Path(f_split[0]).parent), 'Fixed')
    # Crear el nombre del fichero en la ruta 'Fixed' con el mismo nombre que el original pero extensión .xlsx
    filename_out = str(Path(f_split[0]).parent / 'Fixed' / f'{Path(f_split[0]).name}.xlsx')

    # Abrir fichero utilizando codificación 'utf-16'
    fich = io.open(filename, "r", encoding="utf-16")
    data = fich.readlines()

    # Abrir el fichero de salida y crear una página "Data"
    workbook = Workbook()
    sheet = workbook.create_sheet(sheet_name)

    # Iterar sobre los datos y guardarlos en Excel
    for row, fila in enumerate(data):
        # Quitar el '\n' que se inserta cuando se lee con el reader de io.open.
        # Cojer los valores despues de splittear con '\t'
        for column, val in enumerate(fila.replace('\n', '').split('\t')):
            try:
                sheet.cell(row + 1, column + 1, val)
            except openpyxl.utils.exceptions.IllegalCharacterError:
                print(f"Invalid line {row + 1}")
                break
    # Guardar el Excel (se genera)
    workbook.save(filename_out)

    if df_return:
        # Leer el fichero bueno con Pandas
        df_out = pd.read_excel(filename_out, sheet_name=sheet_name, skiprows=skiprows, engine='openpyxl')
    else:
        df_out = None

    return df_out


def generar_excel_default(data, nomfich, hoja="Hoja1", data_col=None, df_general=False, plant=None, nombres_col_gen=None, df_num=None, startrow=0, startcol=0, autofilter=False, index=False, header=True):
    writer = pd.ExcelWriter(nomfich, engine='openpyxl')
    if isinstance(data, np.ndarray):
        if data_col is not None:
            data = pd.DataFrame(data, columns=data_col)
        else:
            raise Exception("With a numpy ndarray is necessary to include data_col")

    if df_num is not None:
        for df in df_num:
            convertir_df_numerico(data, df)

    if plant is not None and nombres_col_gen is not None:
        _cm.convertir_df_numerico_GENERAL(data, nombres_col_gen, plant)

    data.to_excel(writer, sheet_name=hoja, startrow=startrow, startcol=startcol, index=False, header=True)

    # Activar autofilter
    if autofilter:
        worksheet = writer.sheets[hoja]
        worksheet.auto_filter.ref = worksheet.dimensions
        # worksheet.autofilter(0, 0, data.shape[0], data_col.size)  # Se usaría en caso de utilizar xlsxwriter

    writer.save()


def time_calculator(func, *args, **kwargs):
    start = time.time_ns()
    func(*args, **kwargs)
    print(rf"Time: {time.time_ns() - start}")


def find_adjacent_loops(Data, kks_loops, loop, n_adj, front_subfield=True):
    """
    Finds the adjacent loops of a loop in a solar field.

    :param kks_loops: Tags of all the loops at the Solar Field
    :param loop: Target loop
    :param n_adj: Number of adjacent loops to look for. Minimun value is 1 (front loop only would be)

    :param front_subfield: Look for adjacent loops in the subfield that shares with the current subfield. True by default
    :return: Array of 'n_adj' adjacent loops to target loop 'loop'
    """
    if n_adj < 1:
        raise Exception("The number of adjacent loops must be at least 1")

    pos_target = np.argwhere(np.array(np.logical_and(Data.kks_loops_full == loop, Data.pos_loc == 1)))[0, 0]

    # target = plantilla.iloc[pos_target]

    # Numero de adyacentes
    col_target = []
    for k in range(n_adj):
        col_target.append(Data.column[pos_target] + k * 2)
        col_target.append(Data.column[pos_target] - k * 2)

    # Cáclulo de adyacentes
    adj_loops = np.argwhere((Data.sectores_full == Data.sectores_full[pos_target]) &
                            (np.isin(Data.column, col_target)) &
                            (Data.pos_loc == 1)).squeeze(1)

    adj_loops = [np.argwhere(kks_loops == Data.kks_loops_full[loop])[0, 0] for loop in adj_loops]

    return adj_loops


def tag_splitter(tag):
    TAG = Cstruct()
    if hasattr(tag, "__len__"): # Comprueba si tiene el atributo __len__,
        # Array
        pass
    else:
        # String
        TAG.ew_sector = tag[3]


def delta_calculation(data, axis=0, nlocs=4):
    if data.ndim > 1:
        out_data = np.empty([data.shape[0], data.shape[1] - 1], dtype=float)
        for k in range(data.shape[axis]):
            for i in range(nlocs - 1):
                out_data[k, i] = data[k, i] - data[k, i + 1]
    else:
        out_data = np.empty(data.size - 1, dtype=float)
        for i in range(nlocs - 1):
            out_data[i] = data[i] - data[i + 1]

    return out_data


def negate(f):
    @wraps(f)
    def g(*args,**kwargs):
        return not f(*args,**kwargs)
    g.__name__ = f'negate({f.__name__})'
    return g


def guardar_var(ruta, var):  # Guarda una variable en la ruta indicada
    f = open(ruta, 'wb')
    pickle.dump(var, f)
    f.close()


def cargar_var(ruta):  # Devuelve el valor de una variable cargada desde la ruta indicada
    f = open(ruta, 'rb')
    var = pickle.load(f)
    f.close()
    return var



"""
Title: 
Author: Julio Rodriguez
Organization: SENER
"""

# BLK: Imports

# Standard Libraries
import cProfile
import pstats

# 3rd Party

# Local Libraries


# Global variables
class Cstruct:
    pass

# Code

# Clases


class SlowFunction:
    def __init__(self):
        self.profile = cProfile.Profile()

    def start(self):
        self.profile.enable()

    def stop(self):
        self.profile.disable()
        ps = pstats.Stats(self.profile)
        ps.sort_stats('calls', 'cumtime')
        ps.print_stats(20)


# Funciones




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="senerpy",
    version="0.0.6",
    author="Julio Rodriguez",
    author_email="julio.rodriguez@sener.es",
    description="Librerias para uso de Data Science por parte de SENER",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/stulbfl/senerpy",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8",
)